@@ -10,6 +10,8 @@ from http.cookies import SimpleCookie
 from itertools import cycle, islice
 from time import monotonic
 
+import attr
+
 from . import hdrs, helpers
 from .client_exceptions import (ClientConnectionError,
                                 ClientConnectorCertificateError,
@@ -20,7 +22,7 @@ from .client_exceptions import (ClientConnectionError,
                                 ssl_errors)
 from .client_proto import ResponseHandler
 from .client_reqrep import ClientRequest, Fingerprint, _merge_ssl_params
-from .helpers import PY_36, is_ip_address, noop, sentinel
+from .helpers import PY_36, CeilTimeout, is_ip_address, noop, sentinel
 from .locks import EventResultOrError
 from .resolver import DefaultResolver
 
@@ -272,7 +274,8 @@ class BaseConnector:
                     if proto.is_connected():
                         if use_time - deadline < 0:
                             transport = proto.close()
-                            if key[-1] and not self._cleanup_closed_disabled:
+                            if (key.is_ssl and
+                                    not self._cleanup_closed_disabled):
                                 self._cleanup_closed_transports.append(
                                     transport)
                         else:
@@ -388,7 +391,7 @@ class BaseConnector:
 
         return available
 
-    async def connect(self, req, traces=None):
+    async def connect(self, req, traces, timeout):
         """Get from pool or create new connection."""
         key = req.connection_key
         available = self._available_connections(key)
@@ -439,10 +442,7 @@ class BaseConnector:
                     await trace.send_connection_create_start()
 
             try:
-                proto = await self._create_connection(
-                    req,
-                    traces=traces
-                )
+                proto = await self._create_connection(req, traces, timeout)
                 if self._closed:
                     proto.close()
                     raise ClientConnectionError("Connector is closed.")
@@ -482,7 +482,7 @@ class BaseConnector:
                 if t1 - t0 > self._keepalive_timeout:
                     transport = proto.close()
                     # only for SSL transports
-                    if key[-1] and not self._cleanup_closed_disabled:
+                    if key.is_ssl and not self._cleanup_closed_disabled:
                         self._cleanup_closed_transports.append(transport)
                 else:
                     if not conns:
@@ -546,7 +546,7 @@ class BaseConnector:
         if should_close or protocol.should_close:
             transport = protocol.close()
 
-            if key[-1] and not self._cleanup_closed_disabled:
+            if key.is_ssl and not self._cleanup_closed_disabled:
                 self._cleanup_closed_transports.append(transport)
         else:
             conns = self._conns.get(key)
@@ -558,7 +558,7 @@ class BaseConnector:
                 self._cleanup_handle = helpers.weakref_handle(
                     self, '_cleanup', self._keepalive_timeout, self._loop)
 
-    async def _create_connection(self, req, traces=None):
+    async def _create_connection(self, req, traces, timeout):
         raise NotImplementedError()
 
 
@@ -744,21 +744,17 @@ class TCPConnector(BaseConnector):
 
         return self._cached_hosts.next_addrs(key)
 
-    async def _create_connection(self, req, traces=None):
+    async def _create_connection(self, req, traces, timeout):
         """Create connection.
 
         Has same keyword arguments as BaseEventLoop.create_connection.
         """
         if req.proxy:
             _, proto = await self._create_proxy_connection(
-                req,
-                traces=traces
-            )
+                req, traces, timeout)
         else:
             _, proto = await self._create_direct_connection(
-                req,
-                traces=traces
-            )
+                req, traces, timeout)
 
         return proto
 
@@ -818,10 +814,12 @@ class TCPConnector(BaseConnector):
         return None
 
     async def _wrap_create_connection(self, *args,
-                                      req, client_error=ClientConnectorError,
+                                      req, timeout,
+                                      client_error=ClientConnectorError,
                                       **kwargs):
         try:
-            return await self._loop.create_connection(*args, **kwargs)
+            with CeilTimeout(timeout.sock_connect):
+                return await self._loop.create_connection(*args, **kwargs)
         except certificate_errors as exc:
             raise ClientConnectorCertificateError(
                 req.connection_key, exc) from exc
@@ -830,9 +828,8 @@ class TCPConnector(BaseConnector):
         except OSError as exc:
             raise client_error(req.connection_key, exc) from exc
 
-    async def _create_direct_connection(self, req,
-                                        *, client_error=ClientConnectorError,
-                                        traces=None):
+    async def _create_direct_connection(self, req, traces, timeout,
+                                        *, client_error=ClientConnectorError):
         sslcontext = self._get_ssl_context(req)
         fingerprint = self._get_fingerprint(req)
 
@@ -857,7 +854,7 @@ class TCPConnector(BaseConnector):
 
             try:
                 transp, proto = await self._wrap_create_connection(
-                    self._factory, host, port,
+                    self._factory, host, port, timeout=timeout,
                     ssl=sslcontext, family=hinfo['family'],
                     proto=hinfo['proto'], flags=hinfo['flags'],
                     server_hostname=hinfo['hostname'] if sslcontext else None,
@@ -881,7 +878,7 @@ class TCPConnector(BaseConnector):
         else:
             raise last_exc
 
-    async def _create_proxy_connection(self, req, traces=None):
+    async def _create_proxy_connection(self, req, traces, timeout):
         headers = {}
         if req.proxy_headers is not None:
             headers = req.proxy_headers
@@ -896,7 +893,7 @@ class TCPConnector(BaseConnector):
 
         # create connection to proxy server
         transport, proto = await self._create_direct_connection(
-            proxy_req, client_error=ClientProxyConnectionError)
+            proxy_req, [], timeout, client_error=ClientProxyConnectionError)
 
         auth = proxy_req.headers.pop(hdrs.AUTHORIZATION, None)
         if auth is not None:
@@ -918,11 +915,15 @@ class TCPConnector(BaseConnector):
             # asyncio handles this perfectly
             proxy_req.method = hdrs.METH_CONNECT
             proxy_req.url = req.url
-            key = (req.host, req.port, req.ssl)
+            key = attr.evolve(req.connection_key,
+                              proxy=None,
+                              proxy_auth=None,
+                              proxy_headers_hash=None)
             conn = Connection(self, key, proto, self._loop)
             proxy_resp = await proxy_req.send(conn)
             try:
-                resp = await proxy_resp.start(conn, True)
+                conn._protocol.set_response_params()
+                resp = await proxy_resp.start(conn)
             except BaseException:
                 proxy_resp.close()
                 conn.close()
@@ -948,7 +949,8 @@ class TCPConnector(BaseConnector):
                     transport.close()
 
                 transport, proto = await self._wrap_create_connection(
-                    self._factory, ssl=sslcontext, sock=rawsock,
+                    self._factory, timeout=timeout,
+                    ssl=sslcontext, sock=rawsock,
                     server_hostname=req.host,
                     req=req)
             finally:
@@ -981,10 +983,11 @@ class UnixConnector(BaseConnector):
         """Path to unix socket."""
         return self._path
 
-    async def _create_connection(self, req, traces=None):
+    async def _create_connection(self, req, traces, timeout):
         try:
-            _, proto = await self._loop.create_unix_connection(
-                self._factory, self._path)
+            with CeilTimeout(timeout.sock_connect):
+                _, proto = await self._loop.create_unix_connection(
+                    self._factory, self._path)
         except OSError as exc:
             raise ClientConnectorError(req.connection_key, exc) from exc
 
