@@ -1,4 +1,3 @@
-import asyncio
 import socket
 from collections import MutableMapping
 from unittest import mock
@@ -8,11 +7,17 @@ from multidict import CIMultiDict, MultiDict
 from yarl import URL
 
 from aiohttp import HttpVersion
+from aiohttp.helpers import DEBUG
 from aiohttp.streams import StreamReader
 from aiohttp.test_utils import make_mocked_request
 from aiohttp.web import HTTPRequestEntityTooLarge
 
 
+@pytest.fixture
+def protocol():
+    return mock.Mock(_reading_paused=False)
+
+
 def test_ctor():
     req = make_mocked_request('GET', '/path/to?a=1&b=2')
 
@@ -24,7 +29,6 @@ def test_ctor():
     assert 'a=1&b=2' == req.query_string
     assert CIMultiDict() == req.headers
     assert () == req.raw_headers
-    assert req.message == req._message
 
     get = req.query
     assert MultiDict([('a', '1'), ('b', '2')]) == get
@@ -48,8 +52,11 @@ def test_ctor():
     assert req.raw_headers == ((b'FOO', b'bar'),)
     assert req.task is req._task
 
+
+def test_deprecated_message():
+    req = make_mocked_request('GET', '/path/to?a=1&b=2')
     with pytest.warns(DeprecationWarning):
-        assert req.GET is req.query
+        assert req.message == req._message
 
 
 def test_doubleslashes():
@@ -109,6 +116,58 @@ def test_content_length():
     assert 123 == req.content_length
 
 
+def test_range_to_slice_head():
+    def bytes_gen(size):
+        for i in range(size):
+            yield i % 256
+    payload = bytearray(bytes_gen(10000))
+    req = make_mocked_request(
+        'GET', '/',
+        headers=CIMultiDict([('RANGE', 'bytes=0-499')]),
+        payload=payload)
+    assert isinstance(req.http_range, slice)
+    assert req.content[req.http_range] == payload[:500]
+
+
+def test_range_to_slice_mid():
+    def bytes_gen(size):
+        for i in range(size):
+            yield i % 256
+    payload = bytearray(bytes_gen(10000))
+    req = make_mocked_request(
+        'GET', '/',
+        headers=CIMultiDict([('RANGE', 'bytes=500-999')]),
+        payload=payload)
+    assert isinstance(req.http_range, slice)
+    assert req.content[req.http_range] == payload[500:1000]
+
+
+def test_range_to_slice_tail_start():
+    def bytes_gen(size):
+        for i in range(size):
+            yield i % 256
+    payload = bytearray(bytes_gen(10000))
+    req = make_mocked_request(
+        'GET', '/',
+        headers=CIMultiDict([('RANGE', 'bytes=9500-')]),
+        payload=payload)
+    assert isinstance(req.http_range, slice)
+    assert req.content[req.http_range] == payload[-500:]
+
+
+def test_range_to_slice_tail_stop():
+    def bytes_gen(size):
+        for i in range(size):
+            yield i % 256
+    payload = bytearray(bytes_gen(10000))
+    req = make_mocked_request(
+        'GET', '/',
+        headers=CIMultiDict([('RANGE', 'bytes=-500')]),
+        payload=payload)
+    assert isinstance(req.http_range, slice)
+    assert req.content[req.http_range] == payload[-500:]
+
+
 def test_non_keepalive_on_http10():
     req = make_mocked_request('GET', '/', version=HttpVersion(1, 0))
     assert not req.keep_alive
@@ -119,30 +178,27 @@ def test_non_keepalive_on_closing():
     assert not req.keep_alive
 
 
-@asyncio.coroutine
-def test_call_POST_on_GET_request():
+async def test_call_POST_on_GET_request():
     req = make_mocked_request('GET', '/')
 
-    ret = yield from req.post()
+    ret = await req.post()
     assert CIMultiDict() == ret
 
 
-@asyncio.coroutine
-def test_call_POST_on_weird_content_type():
+async def test_call_POST_on_weird_content_type():
     req = make_mocked_request(
         'POST', '/',
         headers=CIMultiDict({'CONTENT-TYPE': 'something/weird'}))
 
-    ret = yield from req.post()
+    ret = await req.post()
     assert CIMultiDict() == ret
 
 
-@asyncio.coroutine
-def test_call_POST_twice():
+async def test_call_POST_twice():
     req = make_mocked_request('GET', '/')
 
-    ret1 = yield from req.post()
-    ret2 = yield from req.post()
+    ret1 = await req.post()
+    ret2 = await req.post()
     assert ret1 is ret2
 
 
@@ -240,6 +296,21 @@ def test_single_forwarded_header():
     assert req.forwarded[0]['proto'] == 'identifier'
 
 
+@pytest.mark.parametrize(
+    "forward_for_in, forward_for_out",
+    [
+        ("1.2.3.4:1234", "1.2.3.4:1234"),
+        ("1.2.3.4", "1.2.3.4"),
+        ('"[2001:db8:cafe::17]:1234"', '[2001:db8:cafe::17]:1234'),
+        ('"[2001:db8:cafe::17]"', '[2001:db8:cafe::17]'),
+    ])
+def test_forwarded_node_identifier(forward_for_in, forward_for_out):
+    header = 'for={}'.format(forward_for_in)
+    req = make_mocked_request('GET', '/',
+                              headers=CIMultiDict({'Forwarded': header}))
+    assert req.forwarded == ({'for': forward_for_out},)
+
+
 def test_single_forwarded_header_camelcase():
     header = 'bY=identifier;fOr=identifier;HOst=identifier;pRoTO=identifier'
     req = make_mocked_request('GET', '/',
@@ -440,34 +511,31 @@ def test_clone_headers_dict():
     assert req2.raw_headers == ((b'B', b'C'),)
 
 
-@asyncio.coroutine
-def test_cannot_clone_after_read(loop):
-    payload = StreamReader(loop=loop)
+async def test_cannot_clone_after_read(loop, protocol):
+    payload = StreamReader(protocol, loop=loop)
     payload.feed_data(b'data')
     payload.feed_eof()
     req = make_mocked_request('GET', '/path', payload=payload)
-    yield from req.read()
+    await req.read()
     with pytest.raises(RuntimeError):
         req.clone()
 
 
-@asyncio.coroutine
-def test_make_too_big_request(loop):
-    payload = StreamReader(loop=loop)
+async def test_make_too_big_request(loop, protocol):
+    payload = StreamReader(protocol, loop=loop)
     large_file = 1024 ** 2 * b'x'
     too_large_file = large_file + b'x'
     payload.feed_data(too_large_file)
     payload.feed_eof()
     req = make_mocked_request('POST', '/', payload=payload)
     with pytest.raises(HTTPRequestEntityTooLarge) as err:
-        yield from req.read()
+        await req.read()
 
     assert err.value.status_code == 413
 
 
-@asyncio.coroutine
-def test_make_too_big_request_adjust_limit(loop):
-    payload = StreamReader(loop=loop)
+async def test_make_too_big_request_adjust_limit(loop, protocol):
+    payload = StreamReader(protocol, loop=loop)
     large_file = 1024 ** 2 * b'x'
     too_large_file = large_file + b'x'
     payload.feed_data(too_large_file)
@@ -475,13 +543,12 @@ def test_make_too_big_request_adjust_limit(loop):
     max_size = 1024**2 + 2
     req = make_mocked_request('POST', '/', payload=payload,
                               client_max_size=max_size)
-    txt = yield from req.read()
+    txt = await req.read()
     assert len(txt) == 1024**2 + 1
 
 
-@asyncio.coroutine
-def test_multipart_formdata(loop):
-    payload = StreamReader(loop=loop)
+async def test_multipart_formdata(loop, protocol):
+    payload = StreamReader(protocol, loop=loop)
     payload.feed_data(b"""-----------------------------326931944431359\r
 Content-Disposition: form-data; name="a"\r
 \r
@@ -497,13 +564,12 @@ d\r
     req = make_mocked_request('POST', '/',
                               headers={'CONTENT-TYPE': content_type},
                               payload=payload)
-    result = yield from req.post()
+    result = await req.post()
     assert dict(result) == {'a': 'b', 'c': 'd'}
 
 
-@asyncio.coroutine
-def test_make_too_big_request_limit_None(loop):
-    payload = StreamReader(loop=loop)
+async def test_make_too_big_request_limit_None(loop, protocol):
+    payload = StreamReader(protocol, loop=loop)
     large_file = 1024 ** 2 * b'x'
     too_large_file = large_file + b'x'
     payload.feed_data(too_large_file)
@@ -511,7 +577,7 @@ def test_make_too_big_request_limit_None(loop):
     max_size = None
     req = make_mocked_request('POST', '/', payload=payload,
                               client_max_size=max_size)
-    txt = yield from req.read()
+    txt = await req.read()
     assert len(txt) == 1024**2 + 1
 
 
@@ -554,3 +620,17 @@ def test_clone_remote():
     req = make_mocked_request('GET', '/')
     req2 = req.clone(remote='11.11.11.11')
     assert req2.remote == '11.11.11.11'
+
+
+@pytest.mark.skipif(not DEBUG,
+                    reason="The check is applied in DEBUG mode only")
+def test_request_custom_attr():
+    req = make_mocked_request('GET', '/')
+    with pytest.warns(DeprecationWarning):
+        req.custom = None
+
+
+def test_remote_with_closed_transport():
+    req = make_mocked_request('GET', '/')
+    req._protocol = None
+    assert req.remote is None
